@@ -28,10 +28,20 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 #include <assert.h>
+#include <fcntl.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <string.h>
+#include <sys/stat.h>
 
+#ifdef _MSC_VER
+#include <io.h>
+#include <share.h>
+#else
+#include <unistd.h>
+#endif
+
+#include <yara/libyara.h>
 #include <yara/utils.h>
 #include <yara/compiler.h>
 #include <yara/exec.h>
@@ -42,6 +52,160 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 #include <yara/strutils.h>
 
 
+static void _yr_compiler_default_include_free(
+    const char* callback_result_ptr,
+    void* user_data)
+{
+  if(callback_result_ptr != NULL)
+  {
+    yr_free((void*)callback_result_ptr);
+  }
+}
+
+
+const char* _yr_compiler_default_include_callback(
+    const char* include_name,
+    const char* calling_rule_filename,
+    const char* calling_rule_namespace,
+    void* user_data)
+{
+  #ifndef _MSC_VER
+  struct stat stbuf;
+  #endif
+
+  #ifdef _MSC_VER
+  char* b = NULL;
+  #endif
+
+  char* s = NULL;
+  char* f;
+  char* file_buffer;
+  char buffer[1024];
+
+  #ifdef _MSC_VER
+  long file_size;
+  #else
+  off_t file_size;
+  #endif
+
+  int fd = -1;
+
+  if (calling_rule_filename != NULL)
+    strlcpy(buffer, calling_rule_filename, sizeof(buffer));
+  else
+    buffer[0] = '\0';
+
+  s = strrchr(buffer, '/');
+
+  #ifdef _MSC_VER
+  b = strrchr(buffer, '\\'); // in Windows both path delimiters are accepted
+  #endif
+
+  #ifdef _MSC_VER
+  if (s != NULL || b != NULL)
+  #else
+  if (s != NULL)
+  #endif
+  {
+    #ifdef _MSC_VER
+    f = (b > s) ? (b + 1) : (s + 1);
+    #else
+    f = s + 1;
+    #endif
+
+    strlcpy(f, include_name, sizeof(buffer) - (f - buffer));
+
+    f = buffer;
+
+    // SECURITY: Potential for directory traversal here.
+    #ifdef _MSC_VER
+    _sopen_s(&fd, f, _O_RDONLY, _SH_DENYRW, _S_IREAD);
+    #else
+    fd = open(f, O_RDONLY);
+    #endif
+
+    // if include file was not found relative to current source file,
+    // try to open it with path as specified by user (maybe user wrote
+    // a full path)
+    if (fd == -1)
+    {
+      f = (char*) include_name;
+    }
+  }
+  else
+  {
+    f = (char*) include_name;
+  }
+
+  if (fd == -1)
+  {
+    // SECURITY: Potential for directory traversal here.
+    #ifdef _MSC_VER
+    _sopen_s(&fd, f, _O_RDONLY, _SH_DENYRW, _S_IREAD);
+    #else
+    fd = open(f, O_RDONLY);
+    #endif
+  }
+
+  if (fd == -1)
+    return NULL;
+
+  #ifdef _MSC_VER
+  file_size = _filelength(fd);
+  if (file_size == -1)
+  {
+    _close(fd);
+    return NULL;
+  }
+  #else
+  if ((fstat(fd, &stbuf) != 0) || (!S_ISREG(stbuf.st_mode)))
+  {
+    close(fd);
+    return NULL;
+  }
+  file_size = stbuf.st_size;
+  #endif
+
+  file_buffer = (char*) yr_malloc(file_size + 1);
+
+  if (file_buffer == NULL)
+  {
+    #ifdef _MSC_VER
+    _close(fd);
+    #else
+    close(fd);
+    #endif
+
+    return NULL;
+  }
+
+  if (file_size != read(fd, file_buffer, file_size))
+  {
+    yr_free(file_buffer);
+
+    #ifdef _MSC_VER
+    _close(fd);
+    #else
+    close(fd);
+    #endif
+
+    return NULL;
+  }
+  else
+  {
+    file_buffer[file_size] = '\0';
+  }
+
+  #ifdef _MSC_VER
+  _close(fd);
+  #else
+  close(fd);
+  #endif
+
+  return file_buffer;
+}
+
+
 YR_API int yr_compiler_create(
     YR_COMPILER** compiler)
 {
@@ -55,14 +219,14 @@ YR_API int yr_compiler_create(
 
   new_compiler->errors = 0;
   new_compiler->callback = NULL;
+  new_compiler->include_callback = _yr_compiler_default_include_callback;
+  new_compiler->include_free = _yr_compiler_default_include_free;
   new_compiler->last_error = ERROR_SUCCESS;
   new_compiler->last_error_line = 0;
   new_compiler->current_line = 0;
   new_compiler->last_result = ERROR_SUCCESS;
-  new_compiler->file_stack_ptr = 0;
   new_compiler->file_name_stack_ptr = 0;
   new_compiler->fixup_stack_head = NULL;
-  new_compiler->allow_includes = 1;
   new_compiler->loop_depth = 0;
   new_compiler->loop_for_of_mem_offset = -1;
   new_compiler->compiled_rules_arena = NULL;
@@ -181,38 +345,18 @@ YR_API void yr_compiler_set_callback(
 }
 
 
-int _yr_compiler_push_file(
+YR_API void yr_compiler_set_include_callback(
     YR_COMPILER* compiler,
-    FILE* fh)
+    YR_COMPILER_INCLUDE_CALLBACK_FUNC include_callback,
+    YR_COMPILER_INCLUDE_FREE_FUNC include_free,
+    void* user_data)
 {
-  if (compiler->file_stack_ptr < MAX_INCLUDE_DEPTH)
-  {
-    compiler->file_stack[compiler->file_stack_ptr] = fh;
-    compiler->file_stack_ptr++;
-    return ERROR_SUCCESS;
-  }
-  else
-  {
-    compiler->last_result = ERROR_INCLUDE_DEPTH_EXCEEDED;
-    return ERROR_INCLUDE_DEPTH_EXCEEDED;
-  }
+  compiler->include_callback = include_callback;
+  compiler->include_free = include_free;
+  compiler->incl_clbk_user_data = user_data;
 }
 
 
-FILE* _yr_compiler_pop_file(
-    YR_COMPILER* compiler)
-{
-  FILE* result = NULL;
-
-  if (compiler->file_stack_ptr > 0)
-  {
-    compiler->file_stack_ptr--;
-    result = compiler->file_stack[compiler->file_stack_ptr];
-  }
-
-  return result;
-}
-
 int _yr_compiler_push_file_name(
     YR_COMPILER* compiler,
     const char* file_name)
@@ -275,7 +419,7 @@ YR_API char* yr_compiler_get_current_file_name(
 }
 
 
-int _yr_compiler_set_namespace(
+static int _yr_compiler_set_namespace(
     YR_COMPILER* compiler,
     const char* namespace_)
 {
@@ -333,6 +477,7 @@ int _yr_compiler_set_namespace(
   return ERROR_SUCCESS;
 }
 
+
 YR_API int yr_compiler_add_file(
     YR_COMPILER* compiler,
     FILE* rules_file,
@@ -344,6 +489,11 @@ YR_API int yr_compiler_add_file(
 
   assert(compiler->compiled_rules_arena == NULL);
 
+  // Don't allow calls to yr_compiler_add_file() if a previous call to
+  // yr_compiler_add_XXXX failed.
+
+  assert(compiler->errors == 0);
+
   if (file_name != NULL)
     _yr_compiler_push_file_name(compiler, file_name);
 
@@ -376,6 +526,11 @@ YR_API int yr_compiler_add_fd(
 
   assert(compiler->compiled_rules_arena == NULL);
 
+  // Don't allow calls to yr_compiler_add_fd() if a previous call to
+  // yr_compiler_add_XXXX failed.
+
+  assert(compiler->last_error == ERROR_SUCCESS);
+
   if (file_name != NULL)
     _yr_compiler_push_file_name(compiler, file_name);
 
@@ -401,11 +556,16 @@ YR_API int yr_compiler_add_string(
     const char* rules_string,
     const char* namespace_)
 {
-  // Don't allow yr_compiler_add_string() after
+  // Don't allow calls to yr_compiler_add_string() after
   // yr_compiler_get_rules() has been called.
 
   assert(compiler->compiled_rules_arena == NULL);
 
+  // Don't allow calls to yr_compiler_add_string() if a previous call to
+  // yr_compiler_add_XXXX failed.
+
+  assert(compiler->last_error == ERROR_SUCCESS);
+
   if (namespace_ != NULL)
     compiler->last_result = _yr_compiler_set_namespace(compiler, namespace_);
   else
@@ -423,7 +583,7 @@ YR_API int yr_compiler_add_string(
 }
 
 
-int _yr_compiler_compile_rules(
+static int _yr_compiler_compile_rules(
   YR_COMPILER* compiler)
 {
   YARA_RULES_FILE_HEADER* rules_file_header = NULL;
@@ -432,14 +592,14 @@ int _yr_compiler_compile_rules(
   YR_EXTERNAL_VARIABLE null_external;
   YR_AC_TABLES tables;
 
-  int8_t halt = OP_HALT;
+  uint8_t halt = OP_HALT;
   int result;
 
   // Write halt instruction at the end of code.
   yr_arena_write_data(
       compiler->code_arena,
       &halt,
-      sizeof(int8_t),
+      sizeof(uint8_t),
       NULL);
 
   // Write a null rule indicating the end.
@@ -599,6 +759,11 @@ YR_API int yr_compiler_get_rules(
   YR_RULES* yara_rules;
   YARA_RULES_FILE_HEADER* rules_file_header;
 
+  // Don't allow calls to yr_compiler_get_rules() if a previous call to
+  // yr_compiler_add_XXXX failed.
+
+  assert(compiler->last_error == ERROR_SUCCESS);
+
   *rules = NULL;
 
   if (compiler->compiled_rules_arena == NULL)
@@ -815,6 +980,8 @@ YR_API char* yr_compiler_get_error_message(
     char* buffer,
     int buffer_size)
 {
+  uint32_t max_strings_per_rule;
+
   switch(compiler->last_error)
   {
     case ERROR_INSUFFICIENT_MEMORY:
@@ -1002,7 +1169,24 @@ YR_API char* yr_compiler_get_error_message(
           buffer,
           buffer_size,
           "regular expression is too complex");
-
+      break;
+    case ERROR_TOO_MANY_STRINGS:
+       yr_get_configuration(
+          YR_CONFIG_MAX_STRINGS_PER_RULE,
+          &max_strings_per_rule);
+       snprintf(
+          buffer,
+          buffer_size,
+          "too many strings in rule \"%s\" (limit: %d)",
+          compiler->last_error_extra_info,
+          max_strings_per_rule);
+      break;
+    case ERROR_INTEGER_OVERFLOW:
+      snprintf(
+          buffer,
+          buffer_size,
+          "integer overflow in \"%s\"",
+          compiler->last_error_extra_info);
   }
 
   return buffer;
