@@ -27,16 +27,15 @@ ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <unistd.h>
+#include <signal.h>
+
 #include <yara.h>
 #include "blob.h"
 #include "util.h"
 
-#if defined(_WIN32) || defined(__CYGWIN__)
-#include <fileapi.h>
-#else
-#include <unistd.h>
-#endif
-#include <fcntl.h>
 
 static void test_boolean_operators()
 {
@@ -189,9 +188,89 @@ static void test_arithmetic_operators()
   assert_true_rule(
       "rule test { condition: 1--1 == 2}", NULL);
 
+  assert_true_rule(
+      "rule test { condition: 2 * -2 == -4}", NULL);
+
+  assert_true_rule(
+      "rule test { condition: -4 * 2 == -8}", NULL);
+
+  assert_true_rule(
+      "rule test { condition: -4 * -4 == 16}", NULL);
+
   assert_true_rule(
       "rule test { condition: -0x01 == -1}", NULL);
 
+  assert_true_rule(
+      "rule test { condition: 0o10 == 8 }", NULL);
+
+  assert_true_rule(
+      "rule test { condition: 0o100 == 64 }", NULL);
+
+  assert_true_rule(
+      "rule test { condition: 0o755 == 493 }", NULL);
+
+  // TODO: This should return ERROR_INTEGER_OVERFLOW, but right now it returns
+  // ERROR_SYNTAX_ERROR because after the lexer aborts with ERROR_INTEGER_OVERFLOW
+  // the parser finds an unexpected end fails with error: unexpected $end.
+  assert_error(
+      "rule test { condition: 9223372036854775808 > 0 }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(
+      "rule test { condition: 9007199254740992KB > 0 }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(  // integer too long
+      "rule test { condition: 8796093022208MB > 0 }",
+      ERROR_SYNTAX_ERROR);
+
+  assert_error(  // integer too long
+     "rule test { condition: 0x8000000000000000 > 0 }",
+     ERROR_SYNTAX_ERROR);
+
+  assert_error(  // integer too long
+     "rule test { condition: 0o1000000000000000000000 > 0 }",
+     ERROR_SYNTAX_ERROR);
+
+  assert_error(
+     "rule test { condition: 0x7FFFFFFFFFFFFFFF + 1 > 0 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: 9223372036854775807 + 1 > 0 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: -9223372036854775807 - 2 > 0 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: -2 + -9223372036854775807 > 0 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: 1 - -9223372036854775807 > 0 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: 0x4000000000000000 * 2 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: 4611686018427387904 * 2 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: 4611686018427387904 * -2 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: -4611686018427387904 * 2 }",
+     ERROR_INTEGER_OVERFLOW);
+
+  assert_error(
+     "rule test { condition: -4611686018427387904 * -2 }",
+     ERROR_INTEGER_OVERFLOW);
 }
 
 
@@ -217,6 +296,14 @@ static void test_bitwise_operators()
       "rule test { condition: 1 << 3 == 8 }",
       NULL);
 
+  assert_true_rule(
+      "rule test { condition: 1 << 64 == 0 }",
+      NULL);
+
+  assert_true_rule(
+      "rule test { condition: 1 >> 64 == 0 }",
+      NULL);
+
   assert_true_rule(
       "rule test { condition: 1 | 3 ^ 3 == 1 | (3 ^ 3) }",
       NULL);
@@ -1185,6 +1272,9 @@ void test_re()
   // Test case for issue #682
   assert_true_regexp("(a|\\b)[a]{1,}", "aaaa", "aaaa");
 
+  // Test for integer overflow in repeat interval
+  assert_regexp_syntax_error("a{2977952116}");
+
   assert_error(
       "rule test { strings: $a = /a\\/ condition: $a }",
       ERROR_SYNTAX_ERROR);
@@ -1204,6 +1294,38 @@ void test_re()
         strings: $a = /MZ.{300,}?t/ \
         condition: !a == 314 }",
       PE32_FILE);
+
+  assert_false_rule(
+      "rule test { strings: $a = /abc[^d]/ nocase condition: $a }",
+      "abcd");
+
+  assert_false_rule(
+      "rule test { strings: $a = /abc[^d]/ condition: $a }",
+      "abcd");
+
+  assert_false_rule(
+      "rule test { strings: $a = /abc[^D]/ nocase condition: $a }",
+      "abcd");
+
+  assert_true_rule(
+      "rule test { strings: $a = /abc[^D]/ condition: $a }",
+      "abcd");
+
+  assert_true_rule(
+      "rule test { strings: $a = /abc[^f]/ nocase condition: $a }",
+      "abcd");
+
+  assert_true_rule(
+      "rule test { strings: $a = /abc[^f]/ condition: $a }",
+      "abcd");
+
+  assert_true_rule(
+      "rule test { strings: $a = /abc[^F]/ nocase condition: $a }",
+      "abcd");
+
+  assert_true_rule(
+       "rule test { strings: $a = /abc[^F]/ condition: $a }",
+       "abcd");
 }
 
 
@@ -1386,6 +1508,13 @@ static void test_modules()
       }",
       NULL);
 
+  assert_true_rule(
+      "import \"tests\" \
+       rule test { \
+        condition: tests.integer_array[256] == 256 \
+      }",
+      NULL);
+
   assert_true_rule(
       "import \"tests\" \
        rule test { \
@@ -1515,6 +1644,16 @@ static void test_modules()
       ERROR_INVALID_MODULE_NAME);
 }
 
+
+static void test_time_module()
+{
+    assert_true_rule(
+        "import \"time\" \
+        rule test { condition: time.now() > 0 }",
+        NULL);
+}
+
+
 #if defined(HASH_MODULE)
 static void test_hash_module()
 {
@@ -1595,58 +1734,70 @@ void test_integer_functions()
 }
 
 
-void test_file_descriptor()
+void test_include_files()
 {
-  YR_COMPILER* compiler = NULL;
-  YR_RULES* rules = NULL;
+  assert_true_rule(
+    "include \"tests/data/true.yar\" rule t { condition: test }",
+    NULL);
+}
 
-#if defined(_WIN32) || defined(__CYGWIN__)
-  HANDLE fd = CreateFile("tests/data/true.yar", GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
-  if (fd == INVALID_HANDLE_VALUE)
+
+void test_process_scan()
+{
+  int pid = fork();
+  int status = 0;
+  int matches = 0;
+  YR_RULES* rules;
+  int rc1, rc2;
+
+  if (pid == 0)
   {
-    fputs("CreateFile failed", stderr);
-    exit(1);
+    /* The string should appear somewhere in the shell's process space. */
+    if (execl("/bin/sh", "/bin/sh", "-c", "VAR='Hello, world!'; sleep 5; true", NULL) == -1)
+      exit(1);
   }
-#else
-  int fd = open("tests/data/true.yar", O_RDONLY);
-  if (fd < 0)
+  assert(pid > 0);
+
+  /* Give child process time to initialize */
+  sleep(1);
+
+  assert( compile_rule("\
+    rule test {\
+      strings:\
+        $a = { 48 65 6c 6c 6f 2c 20 77 6f 72 6c 64 21 }\
+      condition:\
+        all of them\
+    }", &rules) == ERROR_SUCCESS);
+  rc1 = yr_rules_scan_proc(rules, pid, 0, count_matches, &matches, 0);
+  kill(pid, SIGALRM);
+
+  rc2 = waitpid(pid, &status, 0);
+  if (rc2 == -1)
   {
-    perror("open");
+    perror("waitpid");
     exit(EXIT_FAILURE);
   }
-#endif
-  if (yr_compiler_create(&compiler) != ERROR_SUCCESS)
+  if (status != SIGALRM)
   {
-    perror("yr_compiler_create");
+    fprintf(stderr, "Scanned process exited with unexpected status %d\n", status);
     exit(EXIT_FAILURE);
   }
 
-  if (yr_compiler_add_fd(compiler, fd, NULL, NULL) != 0) {
-    perror("yr_compiler_add_fd");
+  switch (rc1) {
+  case ERROR_SUCCESS:
+    if (matches == 0)
+    {
+      fputs("Found no matches\n", stderr);
+      exit(EXIT_FAILURE);
+    }
+    break;
+  case ERROR_COULD_NOT_ATTACH_TO_PROCESS:
+    fputs("Could not attach to process, ignoring this error\n", stderr);
+    break;
+  default:
+    fprintf(stderr, "yr_rules_scan_proc: Got unexpected error %d\n", rc1);
     exit(EXIT_FAILURE);
   }
-
-#if defined(_WIN32) || defined(__CYGWIN__)
-  CloseHandle(fd);
-#else
-  close(fd);
-#endif
-
-  if (yr_compiler_get_rules(compiler, &rules) != ERROR_SUCCESS) {
-    perror("yr_compiler_add_fd");
-    exit(EXIT_FAILURE);
-  }
-
-  if (compiler)
-  {
-    yr_compiler_destroy(compiler);
-  }
-  if (rules)
-  {
-    yr_rules_destroy(rules);
-  }
-
-  return;
 }
 
 
@@ -1673,9 +1824,10 @@ int main(int argc, char** argv)
   test_for();
   test_re();
   test_filesize();
+  test_include_files();
   // test_compile_file();
   // test_compile_files();
-  // test_include_files();
+
   // test_externals();
   // test_callback();
   // test_compare();
@@ -1686,11 +1838,15 @@ int main(int argc, char** argv)
   test_entrypoint();
   test_global_rules();
 
+  #if HAVE_SCAN_PROC_IMPL == 1
+  test_process_scan();
+  #endif
+
   #if defined(HASH_MODULE)
   test_hash_module();
   #endif
 
-  test_file_descriptor();
+  test_time_module();
 
   yr_finalize();
 
